/*
 * Copyright (c) 2013 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/time.h>
#include <errno.h>
#include <netdb.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <rmp440/rmp440.h>

#include "crc16.h"
#include "io.h"

static ssize_t rmpFakeSend(const struct rmp440_io *, const uint8_t *);
static ssize_t rmpFakeReceive(const struct rmp440_io *, uint8_t *, size_t);
static void rmpFakeClose(const struct rmp440_io *);

static struct rmp440_feedback *data;

struct rmp440_io *
rmpFakeInit(const char *filename)
{
	struct rmp440_io *rmp;

	/* Allocate descriptor */
	rmp = (struct rmp440_io *)malloc(sizeof(struct rmp440_io));
	if (rmp == NULL) {
		return NULL;
	}
	data = rmp440FeedbackInit(filename);
	if (data == NULL) {
		free(rmp);
		return NULL;
	}
	/* arbitrary values for items not in the config file */
	data->operational_state = 3;
	data->min_propulsion_batt_soc = 0.95;
	data->aux_batt_soc = 0.75;
	data->yaw_rate_limit = RMP_DEFAULT_MAXIMUM_YAW_RATE;
	data->vel_limit = RMP_DEFAULT_MAXIMUM_VELOCITY;
	data->front_base_batt_1_soc = 0.95;
	data->front_base_batt_2_soc = 0.95;
	data->rear_base_batt_1_soc = 0.95;
	data->rear_base_batt_2_soc = 0.95;
	data->aux_batt_voltage = 76.0;
	data->aux_batt_current = 0.5;
	data->aux_batt_temp = 27;

	rmp->fd = -1;
	rmp->send = rmpFakeSend;
	rmp->receive = rmpFakeReceive;
	rmp->close = rmpFakeClose;
	rmp->interface = RMP440_INTERFACE_FAKE;
	return rmp;
}

static ssize_t
rmpFakeSend(const struct rmp440_io *rmp, const uint8_t *buffer)
{
	uint16_t cmd;
	union {
		uint32_t i;
		float f;
	} v1, v2;

	cmd = buffer[0] << 8 | buffer[1];
	v1.i = buffer[2] << 24 | buffer[3] << 16 | buffer[4] << 8 | buffer[5];
	v2.i = buffer[6] << 24 | buffer[7] << 16 | buffer[8] << 8 | buffer[9];

	switch (cmd) {
	case RMP_CMD_NONE:
		break;
	case RMP_CAN_CMD_ID:
		if (data->operational_state == 4) {
			data->linear_vel = v1.f * data->vel_limit;
			data->differential_wheel_vel = 
			    v2.f * data->yaw_rate_limit;
		} else {
			data->linear_vel = 0.0;
			data->differential_wheel_vel = 0.0;
		}
		break;
	case RMP_CFG_CMD_ID:
		switch (v1.i) {
		case RMP_CMD_SET_OPERATIONAL_MODE:
			switch (v2.i) {
			case RMP_DISABLE_REQUEST:
				data->operational_state = 5;
				break;
			case RMP_POWERDOWN_REQUEST:
				data->operational_state = 5;
				break;
			case RMP_DTZ_REQUEST:
				data->operational_state = 5;
				break;
			case RMP_STANDBY_REQUEST:
				data->operational_state = 3;
				break;
			case RMP_TRACTOR_REQUEST:
				data->operational_state = 4;
				break;
			}
			break;
		case RMP_CMD_SET_MAXIMUM_VELOCITY:
			data->vel_limit = v2.f;
			data->fram_vel_limit = v2.f;
			break;
		case RMP_CMD_SET_MAXIMUM_ACCEL:
			data->fram_accel_limit = v2.f;
			break;
		case RMP_CMD_SET_MAXIMUM_DECEL:
			data->fram_decel_limit = v2.f;
			break;
		case RMP_CMD_SET_MAXIMUM_DTZ_RATE:
			data->fram_dtz_decel_limit = v2.f;
			break;
		case RMP_CMD_SET_COASTDOWN_ACCEL:
			data->fram_coastdown_decel = v2.f;
			break;
		case RMP_CMD_SET_MAXIMUM_TURN_RATE:
			data->yaw_rate_limit = v2.f;
			data->fram_yaw_rate_limit = v2.f;
			break;
		case RMP_CMD_SET_MAXIMUM_TURN_ACCEL:
			data->fram_yaw_accel_limit = v2.f;
			break;
		case RMP_CMD_SET_TIRE_DIAMETER:
			data->fram_tire_diameter = v2.f;
			break;
		case RMP_CMD_SET_WHEEL_BASE_LENGTH:
			data->fram_wheel_base_length = v2.f;
			break;
		case RMP_CMD_SET_WHEEL_TRACK_WIDTH:
			data->fram_wheel_track_width = v2.f;
			break;
		}
	}
	return RMP_NUM_USB_ETH_BYTES;
}

static void
encodeInt(uint32_t i, uint8_t *buf)
{
	buf[0] = (i & 0xff000000) >> 24;
	buf[1] = (i & 0x00ff0000) >> 16;
	buf[2] = (i & 0x0000ff00) >> 8;
	buf[3] = (i & 0x000000ff);
}

static void
encodeFloat(float f, uint8_t *buf)
{
	union {
		int i;
		float f;
	} v;

	v.f = f;
	encodeInt(v.i, buf);
}

static int
feedbackRequested(int i)
{
	int index = i / 32;
	int mask = 1 << (i % 32);

	return data->user_feedback_bitmap[index] & mask;
}

static ssize_t
rmpFakeReceive(const struct rmp440_io *rmp, uint8_t *buffer, size_t len)
{
	int i;
	static float now = 0.0;
	size_t n = 0;
	uint8_t *p = buffer;

	for (i = 0; i < CCU_CHECKSUM_INDEX; i++) {
		if (!feedbackRequested(i))
			continue;
		switch (i) {
		case CCU_FRAME_COUNT_INDEX:
			now += 0.01;
			encodeFloat((float)now, p);
			break;
		case CCU_OPERATIONAL_STATE_INDEX:
			encodeInt(data->operational_state, p);
			break;
		case CCU_MIN_PROP_SOC_INDEX:
			encodeFloat(data->min_propulsion_batt_soc, p);
			break;
		case CCU_AUX_BATT_SOC_INDEX:
			encodeFloat(data->aux_batt_soc, p);
			break;
		case CCU_YAW_RATE_LIMIT_INDEX:
			encodeFloat(data->yaw_rate_limit, p);
			break;
		case CCU_VELOCITY_LIMIT_INDEX:
			encodeFloat(data->vel_limit, p);
			break;
		case CCU_LINEAR_VELOCITY_INDEX:
			encodeFloat(data->linear_vel, p);
			break;
		case CCU_DIFF_WHEEL_VELOCITY_INDEX:
			encodeFloat(data->differential_wheel_vel, p);
			break;
		case CCU_FRAM_VEL_LIMIT_INDEX:
			encodeFloat(data->fram_vel_limit, p);
			break;
		case CCU_FRAM_ACCEL_LIMIT_INDEX:
			encodeFloat(data->fram_accel_limit, p);
			break;
		case CCU_FRAM_DECEL_LIMIT_INDEX:
			encodeFloat(data->fram_decel_limit, p);
			break;
		case CCU_FRAM_MAX_DTZ_DECEL_INDEX:
			encodeFloat(data->fram_dtz_decel_limit, p);
			break;
		case CCU_FRAM_COASTDOWN_ACCEL_INDEX:
			encodeFloat(data->fram_coastdown_decel, p);
			break;
		case CCU_FRAM_YAW_RATE_LIMIT_INDEX:
			encodeFloat(data->fram_yaw_rate_limit, p);
			break;
		case CCU_FRAM_YAW_ACCEL_LIMIT_INDEX:
			encodeFloat(data->fram_yaw_accel_limit, p);
			break;
		case CCU_FRAM_TIRE_DIAMETER_INDEX:
			encodeFloat(data->fram_tire_diameter, p);
			break;
		case CCU_FRAM_WHEEL_BASE_INDEX:
			encodeFloat(data->fram_wheel_base_length, p);
			break;
		case CCU_FRAM_WHEEL_TRACK_WIDTH_INDEX:
			encodeFloat(data->fram_wheel_track_width, p);
			break;
		case CCU_FRAM_TRANSMISSION_RATIO_INDEX:
			encodeFloat(data->fram_transmission_ratio, p);
			break;
		case CCU_FRAM_USER_FEEDBACK_1_INDEX:
			encodeInt(data->user_feedback_bitmap[0], p);
			break;
		case CCU_FRAM_USER_FEEDBACK_2_INDEX:
			encodeInt(data->user_feedback_bitmap[1], p);
			break;
		case CCU_FRAM_USER_FEEDBACK_3_INDEX:
			encodeInt(data->user_feedback_bitmap[2], p);
			break;
		case CCU_FRAM_USER_FEEDBACK_4_INDEX:
			encodeInt(data->user_feedback_bitmap[3], p);
			break;
		default:
			encodeInt(0, p);
			break;
		}
		p += sizeof(uint32_t);
		n += sizeof(uint32_t);
	}
	encodeInt(0, p);
	n += sizeof(uint32_t);
	crc_compute(buffer, n);
	if (n != rmp440ExpectedBytes(data)) {
		fprintf(stderr, "n != expected %lu %lu\n", 
		    n, rmp440ExpectedBytes(data));
		return -1;
	}
	
	return n;
}

static void
rmpFakeClose(const struct rmp440_io *rmp)
{
	free(data);
	return;
}
