/*
 * Copyright (c) 2017 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include <err.h>
#include <errno.h>
#include <netdb.h>
#include <poll.h>
#include <signal.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <rmp440/rmp440.h>

extern void crc_compute(uint8_t *, size_t);
extern void crc_initialize(void);

static struct rmp440_feedback *fb;
static struct sockaddr_storage local_addr;
static socklen_t local_len;

static int
rmp440SimulInit(const char *configFile)
{
	struct addrinfo *res, *res0, hints;
	int fd, error;

	fb = rmp440FeedbackInit(configFile);
	if (fb == NULL)
		err(2, "rmp440FeedbackInit");

	fb->sp_sw_build_id = 20171031;
	fb->operational_state = 3;

	/* Other data, arbitrary values */
	fb->operational_state = 3;
	fb->min_propulsion_batt_soc = 0.95;
	fb->aux_batt_soc = 0.75;
	fb->yaw_rate_limit = RMP_DEFAULT_MAXIMUM_YAW_RATE;
	fb->vel_limit = RMP_DEFAULT_MAXIMUM_VELOCITY;
	fb->front_base_batt_1_soc = 0.95;
	fb->front_base_batt_2_soc = 0.95;
	fb->rear_base_batt_1_soc = 0.95;
	fb->rear_base_batt_2_soc = 0.95;
	fb->aux_batt_voltage = 76.0;
	fb->aux_batt_current = 0.5;
	fb->aux_batt_temp = 27;

	crc_initialize();

	memset(&hints, 0, sizeof(struct addrinfo));
	hints.ai_flags |= AI_PASSIVE;
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_DGRAM;
	hints.ai_protocol = IPPROTO_UDP;
	if ((error = getaddrinfo(NULL, "8080", &hints, &res))) {
		warnx("getaddrinfo: %s", gai_strerror(error));
		return -1;
	}
	res0 = res;
	do {
		if ((fd = socket(res0->ai_family, res0->ai_socktype,
			    res0->ai_protocol)) < 0)
			continue;
		if (bind(fd, (struct sockaddr *)res0->ai_addr,
			res0->ai_addrlen) == 0) {
			memcpy(&local_addr, res0->ai_addr, res0->ai_addrlen);
			local_len = res0->ai_addrlen;
			break;
		}
		close(fd);
		fd = -1;
	} while ((res0 = res0->ai_next) != NULL);
	if (fd != -1) {
		char name[80];
		char port[16];
		getnameinfo(res0->ai_addr, res0->ai_addrlen,
		    name, sizeof(name), port, sizeof(port),
		    NI_NUMERICHOST | NI_NUMERICSERV);
		printf("rmp440 simu listening on %s:%s \n",
		    name, port);
	}
	freeaddrinfo(res);
	return fd;
}

static int
reopen(struct sockaddr *addr, socklen_t len)
{
	int fd;

	if ((fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
		return fd;
	if (bind(fd, addr,len) < 0) {
		close(fd);
		return -1;
	}
	return fd;
}

static float
rmp440SimuDecodeFloat(const uint8_t *val)
{
	union {
		uint32_t i;
		float f;
	} u;

	u.i = ((val[0] << 24) & 0xff000000) |
	    ((val[1] << 16) & 0x00ff0000) |
	    ((val[2] << 8) & 0x0000ff00) |
	    (val[3] & 0x000000ff);
	return u.f;
}

static uint32_t
rmp440SimuDecodeInt(const uint8_t *val)
{
	return ((val[0] << 24) & 0xff000000) |
	    ((val[1] << 16) & 0x00ff0000) |
	    ((val[2] << 8) & 0x0000ff00) |
	    (val[3] & 0x000000ff);
}

static void
rmp440SimuCfg(uint32_t cfg, const uint8_t *buf)
{
	uint32_t i;

	switch (cfg) {
	case RMP_CMD_SET_OPERATIONAL_MODE:
		i = rmp440SimuDecodeInt(buf);
		switch (i) {
		case RMP_DISABLE_REQUEST:
			fb->operational_state = 5;
			break;
		case RMP_POWERDOWN_REQUEST:
			printf("powerdown\n");
			exit(0);
		case RMP_DTZ_REQUEST:
			fb->operational_state = 3;
			break;
		case RMP_STANDBY_REQUEST:
			fb->operational_state = 3;
			break;
		case RMP_TRACTOR_REQUEST:
			fb->operational_state = 4;
			break;
		default:
			printf("illegal operational mode %d\n", i);
			break;
		}
		printf("%s\n",
		    rmp440DecodeOperationalState(fb->operational_state));
		break;
	case RMP_CMD_SET_MAXIMUM_VELOCITY:
		fb->vel_limit = 
		fb->fram_vel_limit = rmp440SimuDecodeFloat(buf);
		break;
	case RMP_CMD_SET_MAXIMUM_ACCEL:
		fb->fram_accel_limit = rmp440SimuDecodeFloat(buf);
		break;
	case RMP_CMD_SET_MAXIMUM_DECEL:
		fb->fram_decel_limit = rmp440SimuDecodeFloat(buf);
		break;
	case RMP_CMD_SET_MAXIMUM_DTZ_RATE:
		fb->fram_dtz_decel_limit = rmp440SimuDecodeFloat(buf);
		break;
	case RMP_CMD_SET_COASTDOWN_ACCEL:
		fb->fram_coastdown_decel = rmp440SimuDecodeFloat(buf);
		break;
	case RMP_CMD_SET_MAXIMUM_TURN_RATE:
		fb->yaw_rate_limit = 
		fb->fram_yaw_rate_limit = rmp440SimuDecodeFloat(buf);
		break;
	case RMP_CMD_SET_MAXIMUM_TURN_ACCEL:
		fb->fram_yaw_accel_limit = rmp440SimuDecodeFloat(buf);
		break;
	case RMP_CMD_SET_TIRE_DIAMETER:
		fb->fram_tire_diameter = rmp440SimuDecodeFloat(buf);
		break;
	case RMP_CMD_SET_WHEEL_BASE_LENGTH:
		fb->fram_wheel_base_length = rmp440SimuDecodeFloat(buf);
		break;
	case RMP_CMD_SET_WHEEL_TRACK_WIDTH:
		fb->fram_wheel_track_width = rmp440SimuDecodeFloat(buf);
		break;
	case RMP_CMD_SET_TRANSMISSION_RATIO:
		fb->fram_transmission_ratio = rmp440SimuDecodeFloat(buf);
		break;
	case RMP_CMD_SET_INPUT_CONFIG_BITMAP:
		fb->fram_config_bitmap = rmp440SimuDecodeInt(buf);
		break;
	case RMP_CMD_SET_USER_FB_1_BITMAP:
		fb->user_feedback_bitmap[0] = rmp440SimuDecodeInt(buf);
		break;
	case RMP_CMD_SET_USER_FB_2_BITMAP:
		fb->user_feedback_bitmap[1] = rmp440SimuDecodeInt(buf);
		break;
	case RMP_CMD_SET_USER_FB_3_BITMAP:
		fb->user_feedback_bitmap[2] = rmp440SimuDecodeInt(buf);
		break;
	case RMP_CMD_SET_USER_FB_4_BITMAP:
		fb->user_feedback_bitmap[3] = rmp440SimuDecodeInt(buf);
		break;
	default:
		printf("unhandled config command\n");
		break;
	}
}

static int
rmp440SimuCmd(int fd, bool *connected)
{
	uint8_t buf[RMP_NUM_USB_ETH_BYTES];
	struct sockaddr_storage src;
	socklen_t srclen = sizeof(struct sockaddr_storage);
	uint32_t cfg;
	uint16_t cmd;
	ssize_t n;
	float vel, yaw;

	n = recvfrom(fd, buf, RMP_NUM_USB_ETH_BYTES, MSG_DONTWAIT,
	    (struct sockaddr *)&src, &srclen);
	if (n != RMP_NUM_USB_ETH_BYTES) {
		warn("SimuCmd: read");
		*connected = false;
		return -1;
	}
	cmd = (buf[0] << 8) + buf[1];
	switch (cmd) {
	case RMP_CAN_CMD_ID:
		vel = rmp440SimuDecodeFloat(buf+2);
		yaw = rmp440SimuDecodeFloat(buf+6);
		fb->vel_target = vel;
		fb->linear_vel = vel;
		fb->yaw_rate_target = yaw;
		printf("rmpMotion %.2f %.2f\n", vel, yaw);

		break;
	case RMP_CFG_CMD_ID:
		cfg = rmp440SimuDecodeInt(buf+2);
		printf("RMP_CFG_CMD %u\n", cfg);
		rmp440SimuCfg(cfg, buf+6);
		break;
	case RMP_CMD_NONE:
		// printf("rmpNone\n");
		break;
	default:
		printf("unhandled cmd %d\n", cmd);
		break;
	}
	if (!*connected) {
		if (connect(fd, (struct sockaddr *)&src, srclen) < 0) {
			warn("SimuCmd: connect");
			*connected = false;
			return -1;
		} else {
			printf("connected\n");
			*connected = true;
		}
	}
	return 0;
}

static void
rmp440EncodeInt(struct rmp440_feedback *fb, uint32_t *ret, int idx)
{
	uint32_t val;

	switch (idx) {
	case CCU_FSW_1_INDEX:
		val = fb->fault_status[0];
		break;
	case CCU_FSW_2_INDEX:
		val = fb->fault_status[1];
		break;
	case CCU_FSW_3_INDEX:
		val = fb->fault_status[2];
		break;
	case CCU_FSW_4_INDEX:
		val = fb->fault_status[3];
		break;
	case CCU_MCU_0_FSW_INDEX:
		val = fb->mcu_fault_status[0];
		break;
	case CCU_MCU_1_FSW_INDEX:
		val = fb->mcu_fault_status[1];
		break;
	case CCU_MCU_2_FSW_INDEX:
		val = fb->mcu_fault_status[2];
		break;
	case CCU_MCU_3_FSW_INDEX:
		val = fb->mcu_fault_status[3];
		break;
	case CCU_OPERATIONAL_STATE_INDEX:
		val = fb->operational_state;
		break;
	case CCU_DYNAMIC_RESPONSE_INDEX:
		val = fb->dynamic_response;
		break;
	case CCU_PSE_DATA_VALID_INDEX:
		val = fb->pse_data_is_valid;
		break;
	case CCU_AUX_BATT_ABB_STATUS_INDEX:
		val = fb->abb_system_status;
		break;
	case CCU_AUX_BATT_STATUS_INDEX:
		val = fb->abb_batt_status;
		break;
	case CCU_AUX_BATT_FAULTS_INDEX:
		val = fb->aux_batt_faults;
		break;
	case CCU_SP_SW_BUILD_ID_INDEX:
		val = fb->sp_sw_build_id;
		break;
	case CCU_UIP_SW_BUILD_ID_INDEX:
		val = fb->uip_sw_build_id;
		break;
	case CCU_FRAM_CFG_BITMAP_INDEX:
		val = fb->fram_config_bitmap;
		break;
	case CCU_FRAM_ETH_PORT_NUMBER_INDEX:
		val = fb->fram_eth_port_number;
		break;
	case CCU_FRAM_USER_FEEDBACK_1_INDEX:
		val = fb->user_feedback_bitmap[0];
		break;
	case CCU_FRAM_USER_FEEDBACK_2_INDEX:
		val = fb->user_feedback_bitmap[1];
		break;
	case CCU_FRAM_USER_FEEDBACK_3_INDEX:
		val = fb->user_feedback_bitmap[2];
		break;
	case CCU_FRAM_USER_FEEDBACK_4_INDEX:
		val = fb->user_feedback_bitmap[3];
		break;
	default:
		warn("rmp440EncodeInt: invalid index %d\n", idx);
		return;
	}
	*ret = val;
}

/*
 * Encode IPv4 addresses as 32 bit values
 *
 * XXX should use sockaddr_in type and properly convert them to/from
 * XXX network order if they need to be used for real.
 */
static void
rmp440EncodeIP(struct rmp440_feedback *fb, uint32_t *ret, int idx)
{
	uint32_t val;

	switch (idx) {
	case CCU_FRAM_ETH_IP_ADDRESS_INDEX:
		val = fb->fram_eth_ip_address;
		break;
	case CCU_FRAM_ETH_SUBNET_MASK_INDEX:
		val = fb->fram_eth_subnet_mask;
		break;
	case CCU_FRAM_ETH_GATEWAY_INDEX:
		val = fb->fram_eth_gateway;
		break;
	default:
		warn("rmp440EncodeIP: invalid index %d\n", idx);
		return;
	}
	*ret = val;
}

static void
rmp440EncodeFloat(struct rmp440_feedback *fb, float *ret, int idx)
{
	float val;

	switch (idx) {
	case CCU_FRAME_COUNT_INDEX:
		val = fb->frame_count;
		break;
	case CCU_MIN_PROP_SOC_INDEX:
		val = fb->min_propulsion_batt_soc;
		break;
	case CCU_AUX_BATT_SOC_INDEX:
		val = fb->aux_batt_soc;
		break;
	case CCU_INERTIAL_X_ACC_INDEX:
		val = fb->inertial_x_acc;
		break;
	case CCU_INERTIAL_Y_ACC_INDEX:
		val = fb->inertial_y_acc;
		break;
	case CCU_INERTIAL_X_RATE_INDEX:
		val = fb->inertial_x_rate;
		break;
	case CCU_INERTIAL_Y_RATE_INDEX:
		val = fb->inertial_y_rate;
		break;
	case CCU_INERTIAL_Z_RATE_INDEX:
		val = fb->inertial_z_rate;
		break;
	case CCU_PSE_PITCH_INDEX:
		val = fb->pse_pitch;
		break;
	case CCU_PSE_PITCH_RATE_INDEX:
		val = fb->pse_pitch_rate;
		break;
	case CCU_PSE_ROLL_INDEX:
		val = fb->pse_roll;
		break;
	case CCU_PSE_ROLL_RATE_INDEX:
		val = fb->pse_roll_rate;
		break;
	case CCU_PSE_YAW_RATE_INDEX:
		val = fb->pse_yaw_rate;
		break;
	case CCU_YAW_RATE_LIMIT_INDEX:
		val = fb->yaw_rate_limit;
		break;
	case CCU_VELOCITY_LIMIT_INDEX:
		val = fb->vel_limit;
		break;
	case CCU_LINEAR_ACCEL_INDEX:
		val = fb->linear_accel;
		break;
	case CCU_LINEAR_VELOCITY_INDEX:
		val = fb->linear_vel;
		break;
	case CCU_DIFF_WHEEL_VELOCITY_INDEX:
		val = fb->differential_wheel_vel;
		break;
	case CCU_RF_WHEEL_VEL_INDEX:
		val = fb->right_front_vel;
		break;
	case CCU_LF_WHEEL_VEL_INDEX:
		val = fb->left_front_vel;
		break;
	case CCU_RR_WHEEL_VEL_INDEX:
		val = fb->right_rear_vel;
		break;
	case CCU_LR_WHEEL_VEL_INDEX:
		val = fb->left_rear_vel;
		break;
	case CCU_RF_WHEEL_POS_INDEX:
		val = fb->right_front_pos;
		break;
	case CCU_LF_WHEEL_POS_INDEX:
		val = fb->left_front_pos;
		break;
	case CCU_RR_WHEEL_POS_INDEX:
		val = fb->right_rear_pos;
		break;
	case CCU_LR_WHEEL_POS_INDEX:
		val = fb->left_rear_pos;
		break;
	case CCU_LINEAR_POS_INDEX:
		val = fb->linear_pos;
		break;
	case CCU_RF_MOTOR_CURRENT_INDEX:
		val = fb->right_front_current;
		break;
	case CCU_LF_MOTOR_CURRENT_INDEX:
		val = fb->left_front_current;
		break;
	case CCU_RR_MOTOR_CURRENT_INDEX:
		val = fb->right_rear_current;
		break;
	case CCU_LR_MOTOR_CURRENT_INDEX:
		val = fb->left_rear_current;
		break;
	case CCU_MAX_MOTOR_CURRENT_INDEX:
		val = fb->max_motor_current;
		break;
	case CCU_RF_MOTOR_CURRENT_LIMIT_INDEX:
		val = fb->right_front_current_limit;
		break;
	case CCU_LF_MOTOR_CURRENT_LIMIT_INDEX:
		val = fb->left_front_current_limit;
		break;
	case CCU_RR_MOTOR_CURRENT_LIMIT_INDEX:
		val = fb->right_rear_current_limit;
		break;
	case CCU_LR_MOTOR_CURRENT_LIMIT_INDEX:
		val = fb->left_rear_current_limit;
		break;
	case CCU_MIN_CURRENT_LIMIT_INDEX:
		val = fb->min_motor_current_limit;
		break;
	case CCU_FPB_1_SOC_INDEX:
		val = fb->front_base_batt_1_soc;
		break;
	case CCU_FPB_2_SOC_INDEX:
		val = fb->front_base_batt_2_soc;
		break;
	case CCU_RPB_1_SOC_INDEX:
		val = fb->rear_base_batt_1_soc;
		break;
	case CCU_RPB_2_SOC_INDEX:
		val = fb->rear_base_batt_2_soc;
		break;
	case CCU_FPB_1_TEMP_INDEX:
		val = fb->front_base_batt_1_temp;
		break;
	case CCU_FPB_2_TEMP_INDEX:
		val = fb->front_base_batt_2_temp;
		break;
	case CCU_RPB_1_TEMP_INDEX:
		val = fb->rear_base_batt_1_temp;
		break;
	case CCU_RPB_2_TEMP_INDEX:
		val = fb->rear_base_batt_2_temp;
		break;
	case CCU_VEL_TARGET_INDEX:
		val = fb->vel_target;
		break;
	case CCU_YAW_TARGET_INDEX:
		val = fb->yaw_rate_target;
		break;
	case CCU_ANGLE_TARGET_INDEX:
		val = fb->angle_target;
		break;
	case CCU_AUX_BATT_VOLTAGE_INDEX:
		val = fb->aux_batt_voltage;
		break;
	case CCU_AUX_BATT_CURRENT_INDEX:
		val = fb->aux_batt_current;
		break;
	case CCU_AUX_BATT_TEMP_INDEX:
		val = fb->aux_batt_temp;
		break;
	case CCU_7P2V_BATTERY_VOLTAGE_INDEX:
		val = fb->ccu_7p2_battery_voltage;
		break;
	case CCU_MCU_0_INST_POWER_INDEX:
		val = fb->mcu_inst_power[0];
		break;
	case CCU_MCU_1_INST_POWER_INDEX:
		val = fb->mcu_inst_power[1];
		break;
	case CCU_MCU_2_INST_POWER_INDEX:
		val = fb->mcu_inst_power[2];
		break;
	case CCU_MCU_3_INST_POWER_INDEX:
		val = fb->mcu_inst_power[3];
		break;
	case CCU_MCU_0_TOTAL_ENERGY_INDEX:
		val = fb->mcu_total_energy[0];
		break;
	case CCU_MCU_1_TOTAL_ENERGY_INDEX:
		val = fb->mcu_total_energy[1];
		break;
	case CCU_MCU_2_TOTAL_ENERGY_INDEX:
		val = fb->mcu_total_energy[2];
		break;
	case CCU_MCU_3_TOTAL_ENERGY_INDEX:
		val = fb->mcu_total_energy[3];
		break;
	case CCU_FRAM_VEL_LIMIT_INDEX:
		val = fb->fram_vel_limit;
		break;
	case CCU_FRAM_ACCEL_LIMIT_INDEX:
		val = fb->fram_accel_limit;
		break;
	case CCU_FRAM_DECEL_LIMIT_INDEX:
		val = fb->fram_decel_limit;
		break;
	case CCU_FRAM_MAX_DTZ_DECEL_INDEX:
		val = fb->fram_dtz_decel_limit;
		break;
	case CCU_FRAM_COASTDOWN_ACCEL_INDEX:
		val = fb->fram_coastdown_decel;
		break;
	case CCU_FRAM_YAW_RATE_LIMIT_INDEX:
		val = fb->fram_yaw_rate_limit;
		break;
	case CCU_FRAM_YAW_ACCEL_LIMIT_INDEX:
		val = fb->fram_yaw_accel_limit;
		break;
	case CCU_FRAM_TIRE_DIAMETER_INDEX:
		val = fb->fram_tire_diameter;
		break;
	case CCU_FRAM_WHEEL_BASE_INDEX:
		val = fb->fram_wheel_base_length;
		break;
	case CCU_FRAM_WHEEL_TRACK_WIDTH_INDEX:
		val = fb->fram_wheel_track_width;
		break;
	case CCU_FRAM_TRANSMISSION_RATIO_INDEX:
		val = fb->fram_transmission_ratio;
		break;
	default:
		warn("rmp440EncodeFloat: invalid index %d\n", idx);
		return;
	}
	*ret = val;
}


/*
 * Masks indicating the type of returned values. Default == unsigned int
 */
static uint32_t rmp_float_feedback_mask[4] = {
	0xFF7FF900, 0x3FFFFFFF, 0x007FFFF2, 0x00000000
};

static uint32_t rmp_ip_feedback_mask[4] = {
	0x0, 0x0, 0x0D000000, 0x0
};


static void
rmp440SimuEncodeVal(struct rmp440_feedback *fb, uint8_t *val, int idx)
{
	union {
		uint32_t i;
		float f;
	} u;
	int i, mask;

	i = idx / 32;
	mask = 1 << (idx % 32);
	u.i = 0;

	if ((rmp_float_feedback_mask[i] & mask) == mask)
		rmp440EncodeFloat(fb, &u.f, idx);
	else if ((rmp_ip_feedback_mask[i] & mask) == mask)
		rmp440EncodeIP(fb, &u.i, idx);
	else
		rmp440EncodeInt(fb, &u.i, idx);

	*val++ = (u.i & 0xff000000) >> 24;
	*val++ = (u.i & 0x00ff0000) >> 16;
	*val++ = (u.i & 0x0000ff00) >> 8;
	*val++ = (u.i & 0x000000ff);
	return;
}

static int
rmp440SimuUpdate(int fd)
{
	ssize_t n, len;
	uint8_t buf[1024], *p;
	int i, j;
	uint32_t mask;

	memset(buf, 0, sizeof(buf));
	p = buf;
	for (i = 0; i < 4; i++) {
		for (j = 0; j < 32; j++) {
			mask = 1<<j;
			if ((fb->user_feedback_bitmap[i] & mask) == mask) {
				rmp440SimuEncodeVal(fb, p, i*32+j);
				p += sizeof(uint32_t);
			}
		}
	}
	fb->frame_count++;
	len = rmp440ExpectedBytes(fb); /* CRC */
	crc_compute(buf, len);
	n = write(fd, buf, len);
	if (n < 0) {
		warn("SimUpdate: write");
		return -1;
	}
	return 0;
}

int
main(int argc, char *argv[])
{
	int fd, n;
	struct pollfd rfd;
	bool connected = false;

	if (argc == 1)
		fd = rmp440SimulInit(NULL);
	else if (argc == 2)
		fd = rmp440SimulInit(argv[1]);
	else
		errx(2, "syntax: rmp4440Simu [configfile]");

	if (fd < 0)
		err(2, "Cannot setup simulation socket");


	while (1) {
		if (fd < 0)
			fd = reopen((struct sockaddr *)&local_addr, local_len);
		if (fd < 0)
			err(2, "socket");
		rfd.fd = fd;
		rfd.events = POLLIN|POLLHUP|POLLERR;
		n = poll(&rfd, 1, 100);
		if (n < 0)
			err(2, "poll");
		if (n > 0) {
			if (rfd.revents & POLLERR) {
				printf("disconnected\n");
				connected = false;
				close(fd);
				fd = -1;
				continue;
			}
			if (rfd.revents & POLLIN) {
				if (rmp440SimuCmd(fd, &connected) < 0) {
					printf("disconnected\n");
					close(fd);
					fd = -1;
				}
			}
		}
		if (connected)
			if (rmp440SimuUpdate(fd) < 0) {
				printf("disconnected\n");
				connected = false;
				close(fd);
				fd = -1;
			}
	} /* while */
	return 0;
}
