/*
 * Copyright (c) 2013 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <rmp440/rmp440.h>

struct rmp440_config *
readConfig(char *device)
{
	struct rmp440_feedback *fb;
	struct rmp440_io *rmp = NULL;
	struct rmp440_config *cfg;

	fb = rmp440FeedbackInit(NULL);
	if (fb == NULL)
		return NULL;
	cfg = (struct rmp440_config *)malloc(sizeof(struct rmp440_config));
	if (cfg == NULL) {
		free(fb);
		return NULL;
	}
	if (device == NULL)
		rmp = rmp440Init(RMP440_INTERFACE_ETHERNET,
		    "10.40.40.40:8080");
	else {
		if (device[0] == '/') /* /dev/ttyACM0 */
			rmp = rmp440Init(RMP440_INTERFACE_USB, device);
		else if (device[0] >= '0' && device[0] <= '9')
			/* ip address */
			rmp = rmp440Init(RMP440_INTERFACE_ETHERNET,
			    device);
	}
	if (rmp == NULL) {
		fprintf(stderr, "Cannot init rmp connection\n");
		free(fb);
		free(cfg);
		return NULL;
	}
	/* send a null command and decode feedback frame */
	rmp440CmdNone(rmp);
	if (rmp440ReceiveAndDecode(rmp, fb) < 0) {
		free(fb);
		free(cfg);
		return NULL;
	}
	
	rmp440ConfigFromFeedback(cfg, fb, rmp);
	
	return cfg;
}

/*
 * Save configuration to nvram 
 */
int
rmp440WriteConfig(const char *device, const char *file)
{
	struct rmp440_io *rmp = NULL;
	struct rmp440_feedback *data;
	struct rmp440_config *cfg;
	in_addr_t addr, mask, gw;

	cfg = rmp440ParseConfigFile(file);
	if (cfg == NULL) {
		fprintf(stderr, "Could not read config %s\n", file);
		return -1;
	}

	data = rmp440FeedbackInit(NULL);
	if (data == NULL)
		return -1;
	if (device[0] == '/') /* /dev/ttyACM0 */
		rmp = rmp440Init(RMP440_INTERFACE_USB, device);
	else if (device[0] >= '0' && device[0] <= '9')
		/* ip address */
		rmp = rmp440Init(RMP440_INTERFACE_ETHERNET,
		    device);
	if (rmp == NULL) {
		fprintf(stderr, "Cannot init rmp connection\n");
		free(data);
		rmp440FreeConfig(cfg);
		return -1;
	}
	rmp440SetMaximumVelocity(rmp, cfg->max_vel);
	rmp440ReceiveAndDecode(rmp, data);
	rmp440SetMaximumAcceleration(rmp, cfg->max_accel);
	rmp440ReceiveAndDecode(rmp, data);
	rmp440SetMaximumDeceleration(rmp, cfg->max_decel);
	rmp440ReceiveAndDecode(rmp, data);
	rmp440SetMaximumDTZDecelRate(rmp, cfg->max_dtz_decel);
	rmp440ReceiveAndDecode(rmp, data);
	rmp440SetCoastDownAccel(rmp, cfg->coastdown_decel);
	rmp440ReceiveAndDecode(rmp, data);
	rmp440SetMaximumTurnRate(rmp, cfg->max_yaw_rate);
	rmp440ReceiveAndDecode(rmp, data);
	rmp440SetMaximumTurnAccel(rmp, cfg->max_yaw_accel);
	rmp440ReceiveAndDecode(rmp, data);
	rmp440SetTireDiameter(rmp, cfg->tire_diameter);
	rmp440ReceiveAndDecode(rmp, data);
	rmp440SetWheelBaseLength(rmp, cfg->wheel_base_length);
	rmp440ReceiveAndDecode(rmp, data);
	rmp440SetWheelTrackWidth(rmp, cfg->wheel_track_width);
	rmp440ReceiveAndDecode(rmp, data);
	rmp440SetTransmissionRatio(rmp, cfg->gearbox_ratio);
	rmp440ReceiveAndDecode(rmp, data);
	rmp440SetInputConfigBitmap(rmp, cfg->config_bitmap);
	rmp440ReceiveAndDecode(rmp, data);
	printf("robot params done\n");
	rmp440SetUserFeedBack(rmp, cfg->feedback_bitmap[0], 0);
	rmp440ReceiveAndDecode(rmp, data);
	rmp440SetUserFeedBack(rmp, cfg->feedback_bitmap[1], 1);
	rmp440ReceiveAndDecode(rmp, data);
	rmp440SetUserFeedBack(rmp, cfg->feedback_bitmap[1], 2);
	rmp440ReceiveAndDecode(rmp, data);
	rmp440SetUserFeedBack(rmp, cfg->feedback_bitmap[1], 3);
	rmp440ReceiveAndDecode(rmp, data);
	printf("user feedback bitmaps done\n");
	addr = cfg->ip_address.s_addr;
	rmp440SetInetAddress(rmp, addr);
	rmp440ReceiveAndDecode(rmp, data);
	rmp440SetInetPort(rmp, cfg->port);
	rmp440ReceiveAndDecode(rmp, data);
	mask = cfg->ip_netmask.s_addr;
	rmp440SetInetMask(rmp, mask);
	rmp440ReceiveAndDecode(rmp, data);
	gw = cfg->ip_gateway.s_addr;
	rmp440SetInetGateway(rmp, gw);
	rmp440ReceiveAndDecode(rmp, data);
	printf("ip params done\n");
	rmp440FreeConfig(cfg);
	rmp440End(rmp);
	free(data);
	return 0;
}

int
main(int argc, char *argv[])
{
	struct rmp440_config *cfg;
	int opt;
	char *file = NULL, *readDev = NULL, *writeDev = NULL;
	FILE *out;

	while ((opt = getopt(argc, argv, "f:r:w:")) != -1) {
		switch (opt) {
		case 'f':
			file = optarg;
			break;
		case 'r':
			readDev = optarg;
			break;
		case 'w':
			writeDev = optarg;
			break;
		default:
			fprintf(stderr, "usage: %s [-f file][-r|-w device]",
			    argv[0]);
			exit(2);
		}
	}
	if (readDev != NULL && writeDev != NULL) {
		fprintf(stderr, "cannot read and write at the same time\n");
		exit(2);
	}
	if (readDev == NULL && writeDev == NULL) {
		if (file == NULL) {
			fprintf(stderr, "no file specified\n");
			exit(2);
		}
		/* display the configuration file */
		cfg = rmp440ParseConfigFile(file);
		if (cfg == NULL) {
			fprintf(stderr, "Could not read config %s\n", file);
			exit(2);
		}
		rmp440PrintConfig(cfg, stdout);
		rmp440FreeConfig(cfg);
		exit(0);
	}
	if (readDev != NULL) {
		cfg = readConfig(readDev);
		if (cfg == NULL) {
			fprintf(stderr, "Could not get configuration\n");
			exit(2);
		}
		if (file == NULL) {
			rmp440PrintConfig(cfg, stdout);
		} else {
			out = fopen(file, "w");
			if (out == NULL) {
				perror(file);
				exit(2);
			}
			rmp440PrintConfig(cfg, out);
			fclose(out);
		}
		rmp440FreeConfig(cfg);
		exit(0);
	}
	if (writeDev != NULL) {
		if (file == NULL) {
			fprintf(stderr, "Need a configuration file\n");
			exit(0);
		}
		if (rmp440WriteConfig(writeDev, file) != 0) {
			fprintf(stderr, "Could not save config to NVRAM\n");
			exit(2);
		}
	}
	return 0;
}
